/*
 * Copyright 2006 Sony Computer Entertainment Inc.
 *
 * Licensed under the MIT Open Source License, for details please see license.txt or the website
 * http://www.opensource.org/licenses/mit-license.php
 *
 */
#include <FX.pch.h> //PCH

#include "cfxData.h"
#include "cfxPass.h"
#include "cfxEffect.h"
					 
COLLADA_(namespace)
{
	namespace FX
	{//-.
//<-----'

FX::Effect::Effect(xs::ID id, CGcontext ctxt)
:NewParamable(nullptr),Id(id),Cg_Context(ctxt)
{
	if(ctxt!=nullptr)
	{
		//create the new effect and set its name
		Cg = cgCreateEffect(ctxt,nullptr,nullptr);
		cgSetEffectName(Cg,Id);
		assert(Cg!=nullptr);
	}
	else assert(0);
}
FX::Effect::~Effect()
{
	//#ifndef SN_TARGET_PS3 //???
	cgDestroyEffect(Cg);
	//#endif

	for(size_t i=0;i<Profiles.size();i++)	
	delete Profiles[i];
	for(size_t i=0;i<Techniques.size();i++)	
	delete Techniques[i];
}
void FX::Effect::Apply()
{
	Annotatable::Apply(this); NewParamable::Apply();

	//////moving technique earlier so the connection can happen before the set because 
	//////propagation of parameter value is not quite implemented properly yet.
	for(size_t i=0;i<Techniques.size();i++)	
	{
		Techniques[i]->Apply();
							 
		//i think there should be a better place for this, but until i know what that is...
		if(!cgValidateTechnique(Techniques[i]->Cg))
		daeEH::Error<<"INVALID TECHNIQUE";
		else daeEH::Verbose<<"VALID TECHNIQUE";
	}
}

FX::Technique::Technique(FX::NewParamable *parent, xs::ID  sid)
:NewParamable(parent),Sid(sid)
{
	Cg = cgCreateTechnique(parent->FindEffect()->Cg,Sid);
}
FX::Technique::~Technique()
{
	for(size_t i=0;i<Passes.size();i++)
	delete Passes[i];	
}
void FX::Technique::Apply()
{
	Annotatable::Apply(this); NewParamable::Apply();

	for(size_t i=0;i<Passes.size();i++)
	Passes[i]->Apply();
}
void FX::Technique::operator()(const daeElement &include__or__code)
{
	//Care should be taken to extract these in order.
	//Not extracting the typeID because of frequency.
	if("code"==include__or__code.getNCName())
	{
		Collada05::const_code code = 
		include__or__code->a<Collada05::const_code>();
		if(code!=nullptr)
		Cg_Code.insert(Cg_Code.end(),
		code->value->begin(),code->value->end());
	}
	else if("include"==include__or__code.getNCName())
	{
		Collada05::const_include include = 
		include__or__code->a<Collada05::const_include>();
		if(include==nullptr) 
		return;

		xs::anyURI URI = include->url; URI.resolve();

		const daeDOM &DOM = *URI.getDOM();
		daeIORequest req(&DOM,nullptr,&URI,&URI);

		#ifdef NDEBUG
		#error Would rather use req.fulfillRequestI.
		#endif
		//This is the system for mapping URIs to memory files.
		daeIOEmpty I; 
		daeIOSecond<> O(req);
		daePlatform &sys = DOM.getPlatform();
		daeRAII::CloseIO _RAII(sys);
		daeIO *IO = _RAII = sys.openIO(I,O);
		if(IO==nullptr) 
		return;  

		size_t size = IO->getLock();
		Cg_Code.resize(Cg_Code.size()+size);
		if(!IO->readIn(&Cg_Code.back()+1-size,size))
		assert(0);
	}
}
	
void FX::Material::SetPassState(int pass)
{
	cgSetPassState(((FX::Technique*)Parent_NewParamable)->Passes[pass]->Cg);
}
void FX::Material::ResetPassState(int pass)
{
	cgResetPassState(((FX::Technique*)Parent_NewParamable)->Passes[pass]->Cg);
}

//-------.
	}//<-'
}

/*C1071*/