/*
 * Copyright 2006 Sony Computer Entertainment Inc.
 *
 * Licensed under the MIT Open Source License, for details please see license.txt or the website
 * http://www.opensource.org/licenses/mit-license.php
 *
 */
#ifndef __COLLADA_RT__CAMERA_H__
#define __COLLADA_RT__CAMERA_H__

#include "Crt/CrtData.h"
#include "Crt/CrtMatrix.h"

COLLADA_(namespace)
{
	namespace RT
	{//-.
//<-----'
		 
/**
 * Previously "InstanceCamera."
 */
class Camera_State
{
COLLADA_(public)
	/**
	 * Node where this instance was instantiated.
	 */
	RT::Stack_Data *Parent;
	/**
	 * The abstract camera where the cam parameters are stored.
	 */
	RT::Camera *Camera;
	/**UNDOCUMENTED	 
	 */
	RT::Float Pan,Tilt,X,Y,Z,Zoom; RT::Matrix Matrix; 

COLLADA_(public)

	Camera_State():Parent(),Camera(),X(),Y(),Z(),Pan(),Tilt(),Zoom()
	{
		RT::MatrixLoadIdentity(Matrix);
	}

	inline void SetPanAndTilt(RT::Float p, RT::Float t)
	{
		Pan+=p; Tilt+=t; Refresh();
	}

	inline void MoveOrbit(RT::Float xx, RT::Float yy)
	{
		RT::MatrixTranslate(Matrix,0,0,-Zoom);
		RT::MatrixTranslate(Matrix,xx,yy,0);
		X = Matrix[M30];
		Y = Matrix[M31];
		Z = Matrix[M32]; Refresh();
	}

	inline void MoveTransform(RT::Float forward, RT::Float strafe, RT::Float jump)
	{
		RT::Float c = cos(Pan*RT::DEGREES_TO_RADIANS);
		RT::Float s = sin(Pan*RT::DEGREES_TO_RADIANS);
		X+=-c*strafe+s*forward;
		Y+=jump;
		Z+=s*strafe+c*forward; Refresh();
	}

	inline void ZoomTransform(RT::Float z)
	{
		Zoom+=z; Refresh();
	}
				 
	void Refresh()
	{
		RT::MatrixLoadIdentity(Matrix);
		RT::MatrixTranslate(Matrix,X,Y,Z);
		RT::MatrixRotateAngleAxis(Matrix,0,1,0,Pan);
		RT::MatrixRotateAngleAxis(Matrix,1,0,0,Tilt);
		RT::MatrixTranslate(Matrix,0,0,Zoom);
	}
};

/**
 * The RT::Camera class holds the parameters for a camera taken from a <camera> inside a <library_cameras>.  
 * These parameters are then used to instantiate an <instance_camera> in a node.  Currently, the
 * instance_cameras reference this data so if you change it here all the related instance_cameras will
 * change too.  This should probably be changed so each <instance_camera> has a copy of the <camera> data,
 * this would allow the parameters of individual <instance_camera> to change.
 */
class Camera : public RT::Base
{
COLLADA_(public)

	RT::Float 
	Xfov, /**< X field of view for perspective camera */
	Yfov, /**< Y field of view for perspective camera */
	Aspect, /**< Aspect Ratio for perspective camera */
	ZNear, /**< Z clip near for perspective camera */
	ZFar, /**< Z clip far for perspective camera */
	Xmag, /**< Nonzero X magnification for an orthographic camera */
	Ymag; /**< Nonzero Y magnification for an orthographic camera */

COLLADA_(public)

	Camera():Xfov(),Yfov(),Aspect(),ZNear(1),ZFar(10000),Xmag(),Ymag(){}
	~Camera(){}

	inline int Refresh()
	{
		int o = Xmag!=0||Ymag!=0?2:3;
		Refresh(o==2?Xmag:Xfov,o==2?Ymag:Yfov); return o;
	}
	inline void Refresh(RT::Float &X, RT::Float &Y)
	{
		if(Aspect==0)
		{
			Aspect = X==0||Y==0?16.0f/9:X/Y; 
		}
		if(Y==0)
		{
			Y = X!=0?Aspect/X:36;
		}
		if(X==0) X = Aspect*Y;
	}
};

//-------.
	}//<-'
}

#endif //__COLLADA_RT__CAMERA_H__
/*C1071*/