/*
 * Copyright 2006 Sony Computer Entertainment Inc.
 *
 * Licensed under the MIT Open Source License, for details please see license.txt or the website
 * http://www.opensource.org/licenses/mit-license.php
 *
 */
#ifndef __COLLADA_RT__DATA_H__
#define __COLLADA_RT__DATA_H__

#include "Crt/CrtMemory.h"
#include "Crt/CrtMatrix.h"

COLLADA_(namespace)
{
	namespace RT
	{//-.
//<-----'

/**
 * This object holds and <up_axis> and <unit> pair.
 * It's compressed, since any node with an <asset>
 * will need a copy, and because it will be 0 most
 * of the time, and this makes comparisons trivial.
 */
struct Asset_Index
{
	unsigned char _i;
	operator size_t()const{ return _i; }
	Asset_Index();
};

/**
 * This is class will be removed when @c daeDatabase
 * replaces @c RT::DBase. The DOM component makes it
 * redundant.
 */
class Base
{
COLLADA_(public)
	
	//SCHEDULED FOR REMOVAL
	daeName Id,Sid; xs::string DocURI;	

	/**
	 * This is currently whatever @c RT::Asset is
	 * when @c RT::Base::Base() is called.
	 */
	const RT::Asset_Index Asset;
	
COLLADA_(public)

	Base():Id(0),Sid(0),DocURI(){}
	virtual ~Base(){}
};
xs::string DocURI(const DAEP::Element *e)
{
	return dae(e)->getDoc()->getDocURI()->data();
}

template<class array> //xs::const_any or <*_array> class
/**
 * This is a new class to try to formalize usage of <accessor>
 * elements.
 * @see @c RT::Accessor<>.
 * @see @c RT::Accessor05.
 * @see @c RT::Accessor08.
 */
class Accessor : public array
{
COLLADA_(public)

	void bind(){ array::operator=(nullptr); }

	const_daeDocumentRef document;
	Collada05::const_accessor accessor;
	Accessor(const daeDocument *doc):document(doc)
	{}
	Accessor(const DAEP::Element *e):document(dae(e)->getDocument())
	{}
	template<class input>
	/**
	 * This avoids confusion between shared/unshared inputs.
	 * @return Returns <input semantic> if the array exists.
	 */
	daeName bind(input &in)
	{
		accessor = document->idLookup
		<Collada05::const_mesh::source>
		(in->source)->technique_common->accessor;
		array::operator=(xs::anyURI(accessor->source->*"",document).get<array>());
		return in->semantic;
	}

	#define COLLADA_RT_ACCESSOR_i_s_iN(x) \
	size_t i = accessor->offset;\
	size_t s = accessor->stride;\
	size_t iN = i+accessor->count*s; assert(s>=x);\
	Adaptor<T> array_value(src_array);
	template<class T> struct Adaptor
	{	
		typedef const daeArray<typename T::__COLLADA__Atom> Array;
		const Array *operator->(){ return &ref; }
		const Array &ref; Adaptor(const T &t):ref(t){}
	};
	void _out_of_range()
	{
		daeEH::Error<<"<accessor> index is out-of-range "<<accessor->source;
	}

	template<class S> void get(S &dst)
	{
		if(*this!=nullptr) 
		get(dst,operator->()->value);
	}
	template<class S, class T>
	void get(std::vector<S> &dst, const T &src_array)
	{
		COLLADA_RT_ACCESSOR_i_s_iN(0)
		dst.reserve(dst.size()+accessor->count);
		for(S tmp;i<iN;i+=s) 		
		if(1==array_value->get1at(i,tmp)) 
		dst.push_back(tmp); else return _out_of_range();
	}

	template<class S> void get16(S &dst)
	{
		Collada05::const_float_array fa =
		dae(*this)->a<Collada05::const_float_array>();
		if(fa!=nullptr) get16(dst,fa->value);
	}
	template<class S, class T>
	void get16(std::vector<S> &dst, const T &src_array)
	{
		COLLADA_RT_ACCESSOR_i_s_iN(16)
		size_t j = dst.size();
		dst.resize(j+accessor->count);		
		for(;i<iN;i+=s,j++)
		{
			RT::Matrix &m = dst[j];
			if(4!=array_value->get4at(j+0,m[M00],m[M01],m[M02],m[M03])
			 ||4!=array_value->get4at(j+4,m[M10],m[M11],m[M12],m[M13])
			 ||4!=array_value->get4at(j+8,m[M20],m[M21],m[M22],m[M23])
			 ||4!=array_value->get4at(j+12,m[M30],m[M31],m[M32],m[M33]))
			{
				dst.resize(j); return _out_of_range();				
			}
			else if(!RT::Matrix_is_COLLADA_order) RT::MatrixTranspose(m);
		}
	}

	template<class S> int get1at(size_t at, S &dst)
	{
		assert(*this!=nullptr);
		return get1at(at,dst,operator->()->value);
	}
	template<class S, class T>
	int get1at(size_t at, S &dst, const T &src_array)
	{
		COLLADA_RT_ACCESSOR_i_s_iN(0)
		if(1==array_value->get1at(i+at*s,dst)) 
		return 1; _out_of_range(); return 0;
	}	
};

//-------.
	}//<-'
}

#endif //__COLLADA_RT__DATA_H__
/*C1071*/