/*
 * Copyright 2006 Sony Computer Entertainment Inc.
 *
 * Licensed under the MIT Open Source License, for details please see license.txt or the website
 * http://www.opensource.org/licenses/mit-license.php
 *
 */
#ifndef __COLLADA_RT__MATRIX_H__
#define __COLLADA_RT__MATRIX_H__

#include "Crt/CrtTypes.h"

COLLADA_(namespace)
{
	namespace RT
	{//-.
//<-----'
		
#ifdef NDEBUG
#error Increase the precision if keeping.
#endif
static const RT::Float
DEGREES_TO_RADIANS = (RT::Float)0.017453292;

inline void MatrixLoadAsset(RT::Matrix &m, RT::Up up, RT::Float meter=1)
{
	const bool x = up==RT::Up::X_UP;
	const bool z = up==RT::Up::Z_UP; const RT::Float l = meter;
	m[M00] = x?0:l;  m[M01] = x?l:0;      m[M02] = 0;      m[M03] = 0;
	m[M10] = x?-l:0; m[M11] = !x&&!z?l:0; m[M12] = z?-l:0; m[M13] = 0;
	m[M10] = 0;      m[M21] = z?l:0;      m[M22] = x?l:0;  m[M23] = 0;
	m[M10] = 0;      m[M31] = 0;          m[M32] = 0;      m[M33] = 1;
}
inline void MatrixLoadIdentity(RT::Matrix &m)
{
	MatrixLoadAsset(m,RT::Up::Y_UP);
}

inline void MatrixCopy(const RT::Matrix &src, RT::Matrix &dst)
{
	memcpy(dst,src,sizeof(RT::Matrix));
}

void Matrix3x4Invert(const RT::Matrix &LM, RT::Matrix &LMI);

void MatrixTranspose(const RT::Matrix &LSrcMtx, RT::Matrix &LDstMtx);

void MatrixTranspose(RT::Matrix &LDstMtx)
{
	RT::Matrix LSrcMtx;
	RT::MatrixCopy(LSrcMtx,LDstMtx);
	RT::MatrixTranspose(LSrcMtx,LDstMtx);
}

void MatrixToQuat(const RT::Matrix &rotMat, RT::Quaternion &RotQuat);

inline void MatrixTransform(RT::Matrix &Matrix, const FX::Float3 &v, FX::Float3 &tv)
{
	tv.x = float(v.x*Matrix[M00]+v.y*Matrix[M10]+v.z*Matrix[M20]+Matrix[M30]);
	tv.y = float(v.x*Matrix[M01]+v.y*Matrix[M11]+v.z*Matrix[M21]+Matrix[M31]);
	tv.z = float(v.x*Matrix[M02]+v.y*Matrix[M12]+v.z*Matrix[M22]+Matrix[M32]);
}

inline void MatrixRotate(RT::Matrix &Matrix, const FX::Float3 &v, FX::Float3 &tv)
{
	tv.x = float(v.x*Matrix[M00]+v.y*Matrix[M10]+v.z*Matrix[M20]);
	tv.y = float(v.x*Matrix[M01]+v.y*Matrix[M11]+v.z*Matrix[M21]);
	tv.z = float(v.x*Matrix[M02]+v.y*Matrix[M12]+v.z*Matrix[M22]);
}

inline void Matrix3x4TransformSkin(RT::Matrix &Matrix, const FX::Float3 &v, FX::Float3 &tv)
{
	tv.x = float(v.x*Matrix[M00]+v.y*Matrix[M10]+v.z*Matrix[M20]+Matrix[M30]);
	tv.y = float(v.x*Matrix[M01]+v.y*Matrix[M11]+v.z*Matrix[M21]+Matrix[M31]);
	tv.z = float(v.x*Matrix[M02]+v.y*Matrix[M12]+v.z*Matrix[M22]+Matrix[M32]);
}
inline void Matrix3x4RotateSkin(RT::Matrix &Matrix, const FX::Float3 &v, FX::Float3 &tv)
{
	tv.x = float(v.x*Matrix[M00]+v.y*Matrix[M10]+v.z*Matrix[M20]);
	tv.y = float(v.x*Matrix[M01]+v.y*Matrix[M11]+v.z*Matrix[M21]);
	tv.z = float(v.x*Matrix[M02]+v.y*Matrix[M12]+v.z*Matrix[M22]);
}

void MatrixRotateAngleAxis(RT::Matrix &LMatrix, RT::Float LAxisX, RT::Float LAxisY, RT::Float LAxisZ, RT::Float LAngle);
																
void MatrixMult(const RT::Matrix &LMtx1, const RT::Matrix &LMtx2, RT::Matrix &LDstMtx);

void MatrixMult(const RT::Matrix &LSrcMtx, RT::Matrix &LDestMtx);

void MatrixTranslate(RT::Matrix &dst, RT::Float x, RT::Float y, RT::Float z);

void MatrixScale(RT::Matrix &dst, RT::Float x, RT::Float y, RT::Float z);

void QuaternionToMatrix(const RT::Quaternion &LQ, RT::Matrix &LMatrix);

////2017: These routines are old code, refactored to merge the many series classes
////into FX::Float1,2,3&4.

//UNUSED/WAS CrtQuat operator*(CrtQuat,CrtQuat).
//2017: Rewriting this UNUSED code so CrtQuat can be retired in favor of FX::Float4.
void QuaternionMult(const RT::Quaternion &q1, const RT::Quaternion &q2, RT::Quaternion &q1_dst)
{
	RT::Float A,B,C,D,E,F,G,H;
	A = (q1[3]+q1[0])*(q2[3]+q2[0]);
	B = (q1[2]-q1[1])*(q2[1]-q2[2]);
	C = (q1[3]-q1[0])*(q2[1]+q2[2]);
	D = (q1[1]+q1[2])*(q2[3]-q2[0]);
	E = (q1[0]+q1[2])*(q2[0]+q2[1]);
	F = (q1[0]-q1[2])*(q2[0]-q2[1]);
	G = (q1[3]+q1[1])*(q2[3]-q2[2]);
	H = (q1[3]-q1[1])*(q2[3]+q2[2]);

	q1_dst[0] = A-(E+F+G+H)*0.5f;
	q1_dst[1] = C+(E-F+G-H)*0.5f;
	q1_dst[2] = D+(E-F-G+H)*0.5f;
	q1_dst[3] = B+(-E-F+G+H)*0.5f; 
}
//UNUSED/WAS CrtQuat::operator*=(). 
//NOTE the sense has been reversed to match MatrixMult().
//2017: Rewriting this UNUSED code so CrtQuat can be retired in favor of FX::Float4.
void QuaternionMult(const RT::Quaternion &q1, RT::Quaternion &q2_and_q1_dst)
{
	//This is admittedly confusing, and untested. Basically it chains.
	RT::QuaternionMult(q1,q2_and_q1_dst,q2_and_q1_dst);
}
//UNUSED/UNTESTED/ORIGINAL CODE
//void CrtQuat::Rotate(const FX::Float3 &axis, float theta);
void QuaternionRotateAngleAxis(RT::Quaternion &dst, const FX::Float3 &axis, float theta)
{
	//If the axis of rotation is (ax, ay, az)- must be a unit vector
	//and the angle is theta (radians)
	theta/=2; //Repurposing this variable.
	dst[3] = cosf(theta); theta = sin(theta); 
	dst[0] = axis.x*theta; dst[1] = axis.y*theta; dst[2] = axis.z*theta; 
}
//UNUSED/REMOVED CODE/PRESERVING ANYWAY
void QuaternionRotateEulerAngles(RT::Quaternion &dst, const FX::Float3 &v)
{
	//reading a rotation or computing a rotation from the animation channels:
	//v is a vector which contains the X, Y and Z euler angles
	RT::Quaternion qX,qY,qZ; 	
	static const float deg2rad = 3.14159f/180; //Was CrtPI.
	RT::QuaternionRotateAngleAxis(qX,FX::Float3(1,0,0),v.x*deg2rad);
	RT::QuaternionRotateAngleAxis(qY,FX::Float3(0,1,0),v.y*deg2rad);	
	RT::QuaternionRotateAngleAxis(qZ,FX::Float3(0,0,1),v.z*deg2rad);
	RT::QuaternionMult(qY,qX,dst); RT::QuaternionMult(qZ,dst);
}

//-------.
	}//<-'
}

#endif //__COLLADA_RT__MATRIX_H__
/*C1071*/