/*
 * Copyright 2006 Sony Computer Entertainment Inc.
 *
 * Licensed under the MIT Open Source License, for details please see license.txt or the website
 * http://www.opensource.org/licenses/mit-license.php
 *
 */
#include <Crt/RT.pch.h> //PCH

#include "Crt/CrtRender.h"
#include "Crt/CrtTexture.h"

COLLADA_(namespace)
{
	namespace RT
	{//-.
//<-----'

void RT::Image::DeleteTexture()
{												
	if(TexId!=0)
	glDeleteTextures(1,&TexId);
	TexId = 0;	
}
void RT::Image::SetTexture()
{
	if(TexId!=0)
	{
		//Enable Texture Mapping
		glEnable(GL_TEXTURE_2D);
		glBindTexture(GL_TEXTURE_2D,TexId);
		assert(TexId!=0);
	}
	else glDisable(GL_TEXTURE_2D);
}

//RT::Frame::Load swap-clears this afterward.
extern std::vector<char> CrtTexture_buffer(0);

//This had been a global API that received a local file name.
//It was only used by this file, and it must be URI based, but
//is request based to imply the URI is document-based & resolved.
static bool CrtTexture_LoadImage(daeIORequest &req, RT::Texture &texObj)
{
	#ifndef IL_VERSION //LEGACY: TGA-ONLY?
	{
		if(!req.localURI->getURI_extensionIs("tga"))
		return false;
	}
	#endif

	#ifdef NDEBUG
	#error Would rather use req.fulfillRequestI.
	#endif
	//This is the system for mapping URIs to memory files.
	daeIOEmpty I; 
	daeIOSecond<daeIOPlugin::Demands::map> O(req);
	daePlatform &sys = RT::Main.DOM.getPlatform();
	//req.resolve(); //Assuming so.
	daeRAII::CloseIO _RAII(sys);
	daeIO *IO = _RAII = sys.openIO(I,O);
	if(IO==nullptr)
	return false;  
	size_t size = IO->getLock();
	if(size==0)
	return false;
	const void *buf = IO->getReadMap();
	if(buf==nullptr)
	{
		//RT::Main.Load is managing this buffer.
		CrtTexture_buffer.resize(size);
		if(!IO->readIn(&CrtTexture_buffer[0],size))
		return false;
		buf = &CrtTexture_buffer[0];
	}	

	#ifndef IL_VERSION //LEGACY: TGA-ONLY?
	{
		return nullptr!=RT::LoadTargaFromMemory(buf,size,&texObj);
	}
	#else //...
	
	//Must this be 0 terminated?
	ILenum type = 
	ilTypeFromExt(req.localURI->getURI_extension().view);
	#ifdef NDEBUG
	#error WHY NOT const void*? TRACE THIS.
	#endif
	if(IL_TYPE_UNKNOWN!=type
	 ||IL_TRUE!=ilIsValidL(type,(void*)buf,size))
	{
		//Maybe ilLoadL(IL_TYPE_UNKNOWN,buf,size); can guess the 
		//type? There doesn't seem to be an API for guessing the
		//type given its data.
		#ifdef NDEBUG
		#error Use IO->readMIME() and <hex format>?
		#endif
		assert(0); return false;
	}
	
	ILuint image;
	ilGenImages(1,&image); ilBindImage(image);
	ilSetData((void*)buf); ilSetInteger(IL_IMAGE_TYPE,type); 

	int bpp = 0;	
	switch(ilGetInteger(IL_IMAGE_FORMAT))
	{
	case IL_LUMINANCE:

		bpp = 1; texObj.Format = GL_LUMINANCE;

	case IL_LUMINANCE_ALPHA: //TODO: Use 2

		bpp = 2; texObj.Format = GL_LUMINANCE_ALPHA;

	case IL_RGBA: case IL_BGRA: 
	
		bpp = 4; texObj.Format = GL_RGBA; break;

	default: bpp = 3; texObj.Format = GL_RGB;
	}

	texObj.Width = ilGetInteger(IL_IMAGE_WIDTH);
	texObj.Height = ilGetInteger(IL_IMAGE_HEIGHT);		

	//There was an error in the old code.
	//ILint paddedwidth = (width+3)&(~3);
	int gl_Width = texObj.Width*bpp/**(CHAR_BIT/8)*/;
	while(0!=(gl_Width&3)) gl_Width++;

	//TODO: Would rather not buffer the data.
	texObj.Data = COLLADA_RT_array_new(char,gl_Width*texObj.Height);
		
	//NOTE: This works because devIL's enums match OpenGL's.
	ilCopyPixels(0,0,0,texObj.Width,texObj.Height,0,texObj.Format,GL_UNSIGNED_BYTE,(void*)texObj.Data);
	
	ilDeleteImage(image); return true;

	#endif //IL_VERSION
}

bool RT::Image::Refresh()
{
	daeURI absURI;
	absURI.setURI_and_resolve(DocURI,URL);

	daeIORequest req(RT::Main.DOM,nullptr,&absURI,&absURI);

	//try to load the file from the path given. If it fails go to the backup		
	if(!CrtTexture_LoadImage(req,*this))
	return false;

	//RT::Main.Stack.CreateTexture(this);
	{
		daeEH::Verbose<<"Creating Texture...";

		if(TexId==0) glGenTextures(1,&TexId);

		//Create Nearest Filtered Texture
		glBindTexture(GL_TEXTURE_2D,TexId);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
		//2017: gluBuild2DMipmaps was used, but mipmap filter is not???
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR/*GL_LINEAR*/);
			
		//There was no power-2 code in the legacy code.
		//This scenario needs to be checked/dealt with.
		assert(Width/2==Width>>1&&Height/2==Height>>1);

		//2017: Remove GLU dependency.
		//This API calls glTexImage2D.
		//gluBuild2DMipmaps(GL_TEXTURE_2D,Format,Width,Height,Format,GL_UNSIGNED_BYTE,Data);
		glTexImage2D(GL_TEXTURE_2D,0,Format,Width,Height,0,Format,GL_UNSIGNED_BYTE,Data);
		//Note: cfxSampler.cpp does this according to the min-filter.
		//(So this--while harmless--may not be right place for this.)
		#ifdef SN_TARGET_PS3
		glGenerateMipmapOES(GL_TEXTURE_2D);
		#else 		
		glGenerateMipmap(GL_TEXTURE_2D); //Or glGenerateMipmapEXT?
		#endif

		daeEH::Verbose<<"Done."; //return true;
	}

	//2017: There's no sense keeping a copy since glTexImage2D does.
	//(The legacy code that refers to RT::Texture may use its Data.)
	COLLADA_RT_array_delete(Data); return true;	
}

//SCHEDULED FOR REMOVAL?
RT::Texture *RT::LoadTargaFromURI(const daeURI &URI)
{
	URI.resolve();
	daeIORequest req(RT::Main.DOM,nullptr,&URI,&URI);

	RT::Texture tmp;
	if(!CrtTexture_LoadImage(req,tmp))
	return nullptr;

	RT::Texture *out = COLLADA_RT_new(RT::Texture);

	*out = tmp; tmp.Data = nullptr; return out;
}

bool RT::Resource::Locate(const xs::anyURI &URI)
{
	Locate(); 
	
	URI.resolve();
	daeIORequest req(URI.getDOM(),nullptr,&URI,&URI);

	//This is the system for mapping URIs to memory files.
	daeIOEmpty I; 
	daeIOSecond<> O(req);
	daePlatform &sys = RT::Main.DOM.getPlatform();
	//req.resolve(); //Assuming so.
	daeRAII::CloseIO _RAII(sys);
	daeIO *IO = _RAII = sys.openIO(I,O);
	if(IO==nullptr)
	return false;  
	Size = IO->getLock();
	if(Size==0)
	return false;
	Data = COLLADA_RT_array_new(char,Size+1);
	if(!IO->readIn((void*)Data,Size))
	{
		Locate();
		return false;
	}	
	const_cast<char*&>(Data)[Size-1] = '\0';
	daeEH::Verbose<<"Loaded URI "<<URI.getURI()<<".\n"
	"(Size is "<<Size<<".)";	
	return true;
}
  
//-------.
	}//<-'
}

/*C1071*/